""" opentrons_shared_data.module: functions and types for module defs """

import json
from ast import literal_eval
from pathlib import Path
from typing import Union, cast, overload
from functools import lru_cache
from itertools import product

from ..load import load_shared_data
from .types import (
    SchemaVersions,
    ModuleSchema,
    SchemaV1,
    SchemaV3,
    ModuleDefinitionV1,
    ModuleDefinitionV3,
    ModuleModel,
    TOFSensorBaseline,
)


OLD_TC_GEN2_LABWARE_OFFSET = {"x": 0, "y": 68.06, "z": 98.26}

# TODO (spp, 2023-02-14): these values are measured experimentally, and aren't from
#  machine drawings. We should replace them with values from CAD files and
#  possibly make them a part of thermocycler/ deck definitions
FLEX_TC_LID_CLIP_POSITIONS_IN_DECK_COORDINATES = {
    "left_clip": {"x": -3.25, "y": 402, "z": 205},
    "right_clip": {"x": 97.75, "y": 402, "z": 205},
}
FLEX_TC_LID_COLLISION_ZONE = {
    "back_left": {"x": -43.25, "y": 454.9, "z": 211.91},
    "front_right": {"x": 128.75, "y": 402, "z": 211.91},
}
"""
Deck co-ordinates of the top plane of TC lid + lid clips
of a thermocycler on a Flex.
"""


# TODO (spp, 2022-05-12): Python has a built-in error called `ModuleNotFoundError` so,
#                         maybe rename this one?


class ModuleNotFoundError(KeyError):
    def __init__(self, version: str, model_or_loadname: str) -> None:
        super().__init__(model_or_loadname)
        self.requested_version = version
        self.requested_module = model_or_loadname

    def __str__(self) -> str:
        return (
            f"No such version {self.requested_version} module "
            f"{self.requested_module}"
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}: {self.requested_module} "
            f"at version {self.requested_version}"
        )


def load_schema(version: SchemaVersions) -> ModuleSchema:
    path = Path("module") / "schemas" / f"{version}.json"
    return cast(ModuleSchema, json.loads(load_shared_data(path)))


@overload
def load_definition(version: SchemaV1, model_or_loadname: str) -> ModuleDefinitionV1:
    ...


@overload
def load_definition(
    version: SchemaV3,
    model_or_loadname: Union[str, ModuleModel],
) -> ModuleDefinitionV3:
    ...


def load_definition(
    version: SchemaVersions,
    model_or_loadname: Union[str, ModuleModel],
) -> Union[ModuleDefinitionV1, ModuleDefinitionV3]:
    if version == "1":
        path = Path("module") / "definitions" / "1.json"
        data = json.loads(load_shared_data(path))
        try:
            return cast(ModuleDefinitionV1, data[model_or_loadname])
        except KeyError:
            raise ModuleNotFoundError("1", model_or_loadname)
    else:
        path = Path(f"module/definitions/{version}/{model_or_loadname}.json")
        try:
            data = load_shared_data(path)
        except FileNotFoundError:
            raise ModuleNotFoundError(version, model_or_loadname)
        return cast(ModuleDefinitionV3, json.loads(data))


@lru_cache
def load_tof_baseline_data(
    model_or_loadname: Union[str, ModuleModel],
) -> TOFSensorBaseline:
    try:
        definition = load_definition("3", model_or_loadname)
        baseline = definition.get("uniqueModuleData", {})["TOFSensorBaseline"]
        baseline["version"] = baseline.get("version", 1)
        # The baseline is stored as string, so convert to dict
        for axis, platform in product(["X", "Z"], ["extend", "retract"]):
            values = literal_eval(baseline[axis][platform])
            baseline[axis][platform] = values
        return cast(TOFSensorBaseline, baseline)
    except (KeyError, ValueError):
        raise ModuleNotFoundError("3", model_or_loadname)



from __future__ import annotations

"""
opentrons_shared_data.pipette: functions and types for pipette config
"""
import copy
from typing import TYPE_CHECKING, Dict, Optional
import json
from functools import lru_cache

from .. import load_shared_data

if TYPE_CHECKING:
    from .types import (
        PipetteNameSpecs,
        PipetteModelSpecs,
        PipetteName,
        PipetteModel,
        PipetteFusedSpec,
        ChannelCount,
    )


# TODO (spp, 2023-06-22: should probably move this to definition)
"""
The span of pipettes in X-direction based on number of channels.
This is needed in order to determine safe tip drop location within a labware.
"""
PIPETTE_X_SPAN: Dict[ChannelCount, float] = {
    1: 75,  # includes a margin
    8: 75,  # includes a margin
    96: 161,
}


def model_config() -> PipetteModelSpecs:
    """Load the per-pipette-model config file from within the wheel"""
    return copy.deepcopy(_model_config())


@lru_cache(maxsize=1)
def _model_config() -> PipetteModelSpecs:
    return json.loads(
        load_shared_data("pipette/definitions/1/pipetteModelSpecs.json") or "{}"
    )


def name_config() -> PipetteNameSpecs:
    """Load the per-pipette-name config file from within the wheel"""
    return _name_config()


@lru_cache(maxsize=1)
def _name_config() -> PipetteNameSpecs:
    return json.loads(
        load_shared_data("pipette/definitions/1/pipetteNameSpecs.json") or "{}"
    )


def name_for_model(pipette_model: PipetteModel) -> PipetteName:
    """Quickly look up the name for this model"""
    return model_config()["config"][pipette_model]["name"]


def fuse_specs(
    pipette_model: PipetteModel, pipette_name: Optional[PipetteName] = None
) -> PipetteFusedSpec:
    """Combine the model and name spec for a given model.

    if pipette_name is not given, the name field of the pipette config
    is used. If it is, the given name must be in the backCompatNames field.
    """
    return copy.deepcopy(_fuse_specs_cached(pipette_model, pipette_name))


@lru_cache(maxsize=10)
def _fuse_specs_cached(
    pipette_model: PipetteModel, pipette_name: Optional[PipetteName] = None
) -> PipetteFusedSpec:
    """
    Do the work of fusing the specs inside an lru cache. This can't be the
    function that's directly called because we want to return a new object
    all the time, hence the wrapper.
    """
    model_data = _model_config()["config"][pipette_model]
    pipette_name = pipette_name or model_data["name"]

    valid_names = [model_data["name"]] + model_data.get("backCompatNames", [])

    if pipette_name not in valid_names:
        raise KeyError(
            f"pipette name {pipette_name} is not valid for model " f"{pipette_model}"
        )
    name_data = _name_config()[pipette_name]
    # unfortunately, mypy can't verify this way to build typed dicts - we'll
    # make sure it's correct in the tests, and leave the function annotated
    # properly
    return {**model_data, **name_data}


def dummy_model_for_name(pipette_name: PipetteName) -> PipetteModel:
    if "gen2" in pipette_name:
        return "_".join(pipette_name.split("_")[:-1]) + "_v2.0"  # type: ignore
    elif "flex" in pipette_name:
        return "_".join(pipette_name.split("_")[:-1]) + "_v3.0"  # type: ignore
    else:
        return pipette_name + "_v1"  # type: ignore



import json
import pytest
from typing import Any, Dict
from pathlib import Path

from opentrons_shared_data import load_shared_data
from opentrons_shared_data.protocol.models import protocol_schema_v7

from . import list_fixtures


@pytest.mark.parametrize("defpath", list_fixtures(7))
def test_v7_types(defpath: Path) -> None:
    def_data = load_shared_data(defpath)
    def_model = protocol_schema_v7.ProtocolSchemaV7.model_validate_json(def_data)
    def_dict_from_model = def_model.model_dump(
        exclude_unset=True,
        # 'schemaVersion' in python is '$schemaVersion' in JSON
        by_alias=True,
    )
    expected_def_dict = json.loads(def_data)
    delete_unexpected_results(expected_def_dict)
    assert def_dict_from_model == expected_def_dict


# TODO (Tamar and Max 3/16/22):
# Some of our v7 fixtures accidentally contain command "result" data,
# which is not part of the schema, and which we don't parse into our models.
# Here, we delete the extra junk from the fixtures.
# Remove this when the fixtures are corrected.
# https://github.com/Opentrons/opentrons/issues/9701
def delete_unexpected_results(protocol_fixture: Dict[str, Any]) -> None:
    for command_object_dict in protocol_fixture["commands"]:
        command_object_dict.pop("result", None)
        command_object_dict.pop("id", None)



import json
import pytest

from opentrons_shared_data.module import (
    load_definition,
    load_schema,
    ModuleNotFoundError,
)
from opentrons_shared_data import load_shared_data

from . import list_v2_defs, list_v3_defs


@pytest.mark.parametrize("def_name", list_v3_defs())
def test_load_v3_defs(def_name: str) -> None:
    """Test that all v3 definitions load correctly."""
    assert load_definition("3", def_name) == json.loads(
        load_shared_data(f"module/definitions/3/{def_name}.json")
    )


@pytest.mark.parametrize("def_name", list_v2_defs())
def test_load_v2_defs(def_name: str) -> None:
    assert load_definition("2", def_name) == json.loads(  # type: ignore [call-overload]
        load_shared_data(f"module/definitions/2/{def_name}.json")
    )


@pytest.mark.parametrize("def_name", ["magdeck", "tempdeck", "thermocycler"])
def test_load_v1_defs(def_name: str) -> None:
    assert (
        load_definition("1", def_name)
        == json.loads(load_shared_data("module/definitions/1.json"))[def_name]
    )


def test_bad_module_name_throws() -> None:
    with pytest.raises(ModuleNotFoundError):
        load_definition("1", "alsjdag")

    with pytest.raises(ModuleNotFoundError):
        load_definition("2", "asdasda")  # type: ignore[call-overload]


@pytest.mark.parametrize("schema_name", ["1", "2", "3"])
def test_load_schema(schema_name: str) -> None:
    assert load_schema(schema_name) == json.loads(  # type: ignore[arg-type]
        load_shared_data(f"module/schemas/{schema_name}.json")
    )



"""
opentrons_shared_data.deck.types: types for deck defs

This should only be imported if typing.TYPE_CHECKING is True
"""

from typing import Any, Dict, List, NewType, Union
from typing_extensions import Literal, TypedDict

from ..module.types import ModuleType
from opentrons_shared_data.labware.types import LocatingFeatures


DeckSchemaVersion5 = Literal[5]
DeckSchemaVersion4 = Literal[4]
DeckSchemaVersion3 = Literal[3]
DeckSchemaVersion2 = Literal[2]
DeckSchemaVersion1 = Literal[1]

DeckSchema = NewType("DeckSchema", Dict[str, Any])

DeckSchemaId = Literal["opentronsDeckSchemaV3", "opentronsDeckSchemaV4"]
RobotModel = Union[Literal["OT-2 Standard"], Literal["OT-3 Standard"]]


class Metadata(TypedDict, total=False):
    displayName: str
    tags: List[str]


class Robot(TypedDict):
    model: RobotModel


class BoundingBox(TypedDict):
    xDimension: float
    yDimension: float
    zDimension: float


class SlotDefV3(TypedDict, total=False):
    id: str
    position: List[float]
    boundingBox: BoundingBox
    displayName: str
    compatibleModuleTypes: List[ModuleType]
    matingSurfaceUnitVector: List[Union[Literal[1], Literal[-1]]]
    features: LocatingFeatures


class CalibrationPoint(TypedDict):
    id: str
    position: List[float]
    displayName: str


class INode(TypedDict):
    name: str
    type: str
    value: str
    attributes: Dict[str, str]
    # this should be a recursive call to INode but we need to upgrade mypy
    children: List[Dict[str, Any]]


class FixedLabwareBySlot(TypedDict):
    id: str
    displayName: str
    labware: str
    slot: str


class FixedLabwareByPosition(TypedDict):
    id: str
    displayName: str
    labware: str
    position: List[float]


class FixedVolumeBySlot(TypedDict):
    id: str
    displayName: str
    boundingBox: BoundingBox
    slot: str


class FixedVolumeByPosition(TypedDict):
    id: str
    displayName: str
    boundingBox: BoundingBox
    position: List[float]


class _RequiredAddressableArea(TypedDict):
    id: str
    areaType: str
    offsetFromCutoutFixture: List[float]
    boundingBox: BoundingBox
    displayName: str


class AddressableArea(_RequiredAddressableArea, total=False):
    compatibleModuleTypes: List[ModuleType]
    matingSurfaceUnitVector: List[Union[Literal[1], Literal[-1]]]
    ableToDropTips: bool
    ableToDropLabware: bool


class AddressableAreaV5(_RequiredAddressableArea, total=False):
    compatibleModuleTypes: List[ModuleType]
    matingSurfaceUnitVector: List[Union[Literal[1], Literal[-1]]]
    ableToDropTips: bool
    ableToDropLabware: bool
    features: LocatingFeatures


class Cutout(TypedDict):
    id: str
    position: List[float]
    displayName: str


class CutoutFixture(TypedDict):
    id: str
    expectOpentronsModuleSerialNumber: bool
    mayMountTo: List[str]
    displayName: str
    providesAddressableAreas: Dict[str, List[str]]
    fixtureGroup: Dict[str, List[Dict[str, str]]]
    height: float


Fixture = Union[
    FixedLabwareBySlot, FixedLabwareByPosition, FixedVolumeBySlot, FixedVolumeByPosition
]


class LocationsV3(TypedDict):
    orderedSlots: List[SlotDefV3]
    calibrationPoints: List[CalibrationPoint]
    fixtures: List[Fixture]


class LocationsV4(TypedDict):
    addressableAreas: List[AddressableArea]
    calibrationPoints: List[CalibrationPoint]
    cutouts: List[Cutout]
    legacyFixtures: List[Fixture]


class LocationsV5(TypedDict):
    addressableAreas: List[AddressableAreaV5]
    calibrationPoints: List[CalibrationPoint]
    cutouts: List[Cutout]
    legacyFixtures: List[Fixture]


class NamedOffset(TypedDict):
    x: float
    y: float
    z: float


class GripperOffsets(TypedDict):
    pickUpOffset: NamedOffset
    dropOffset: NamedOffset


class _RequiredDeckDefinitionV3(TypedDict):
    otId: str
    schemaVersion: Literal[3]
    cornerOffsetFromOrigin: List[float]
    dimensions: List[float]
    metadata: Metadata
    robot: Robot
    locations: LocationsV3
    layers: List[INode]


class DeckDefinitionV3(_RequiredDeckDefinitionV3, total=False):
    gripperOffsets: Dict[str, GripperOffsets]


class _RequiredDeckDefinitionV4(TypedDict):
    otId: str
    schemaVersion: Literal[4]
    cornerOffsetFromOrigin: List[float]
    dimensions: List[float]
    metadata: Metadata
    robot: Robot
    locations: LocationsV4
    cutoutFixtures: List[CutoutFixture]


class DeckDefinitionV4(_RequiredDeckDefinitionV4, total=False):
    gripperOffsets: Dict[str, GripperOffsets]


class _RequiredDeckDefinitionV5(TypedDict):
    otId: str
    schemaVersion: Literal[5]
    cornerOffsetFromOrigin: List[float]
    dimensions: List[float]
    metadata: Metadata
    robot: Robot
    locations: LocationsV5
    cutoutFixtures: List[CutoutFixture]


class DeckDefinitionV5(_RequiredDeckDefinitionV5, total=False):
    gripperOffsets: Dict[str, GripperOffsets]


DeckDefinition = Union[DeckDefinitionV3, DeckDefinitionV4, DeckDefinitionV5]



import logging
import json
import re
from pathlib import Path
from typing import Optional, List, Dict, Any, cast
from enum import Enum

from .pipette_definition import PipetteConfigurations, PipetteModelVersionType
from .model_constants import (
    MUTABLE_CONFIGS_V1,
    VALID_QUIRKS,
    _MAP_KEY_TO_V2,
    _MIN_MAX_LOOKUP,
    _TYPE_LOOKUP,
    _UNITS_LOOKUP,
)
from .load_data import load_definition, load_serial_lookup_table
from .types import (
    MutableConfig,
    Quirks,
    QuirkConfig,
    TypeOverrides,
    OverrideType,
    LiquidClasses,
)
from .pipette_load_name_conversions import (
    convert_pipette_model,
    convert_to_pipette_name_type,
)
from .file_operation_helpers import (
    MutableConfigurationEncoder,
    MutableConfigurationDecoder,
)
from .types import PipetteModel, PipetteName


log = logging.getLogger(__name__)

PIPETTE_SERIAL_MODEL_LOOKUP = load_serial_lookup_table()
SERIAL_STUB_REGEX = re.compile(r"P[0-9]{1,3}[KSMHV]{1,2}V[0-9]{2}")
LIQUID_CLASS = LiquidClasses.default


def _edit_non_quirk(  # noqa: C901
    mutable_config_key: str, new_mutable_value: MutableConfig, base_dict: Dict[str, Any]
) -> None:
    def _do_edit_non_quirk(
        new_value: MutableConfig, existing: Dict[Any, Any], keypath: List[Any]
    ) -> None:
        thiskey: Any = keypath[0]
        if thiskey in [lc.name for lc in LiquidClasses]:
            thiskey = LiquidClasses[thiskey]
        if len(keypath) > 1:
            restkeys = keypath[1:]
            if thiskey == "##EACHNOZZLEMAP##":
                for key in existing.keys():
                    _do_edit_non_quirk(new_value, existing[key], restkeys)
            elif thiskey == "##EACHTIPTYPE##":
                for key in existing.keys():
                    _do_edit_non_quirk(new_value, existing[key], restkeys)
            elif thiskey == "##EACHTIP##":
                for key in existing.keys():
                    _do_edit_non_quirk(new_value, existing[key], restkeys)
            else:
                _do_edit_non_quirk(new_value, existing[thiskey], restkeys)
        else:
            # This was the last key
            existing[thiskey] = new_value.value

    new_names = _MAP_KEY_TO_V2[mutable_config_key]
    _do_edit_non_quirk(new_mutable_value, base_dict, new_names)


def _migrate_to_v2_configurations(
    base_configurations: PipetteConfigurations,
    v1_mutable_configs: OverrideType,
) -> PipetteConfigurations:
    """Helper function to migration V1 Configs to the V2 format.

    Given an input of v1 mutable configs, look up the equivalent keyed
    value of that configuration."""
    quirks_list = []
    dict_of_base_model = base_configurations.model_dump(by_alias=True)
    for c, v in v1_mutable_configs.items():
        if isinstance(v, str):
            # ignore the saved model
            continue
        if isinstance(v, bool):
            # ignore any accidental top level quirks.
            continue
        if c == "quirks" and isinstance(v, dict):
            quirks_list.extend([b.name for b in v.values() if b.value])
        elif isinstance(v, MutableConfig):
            _edit_non_quirk(c, v, dict_of_base_model)

    dict_of_base_model["quirks"] = list(
        set(dict_of_base_model["quirks"]).union(set(quirks_list))
    )

    # re-serialization is not great for this nested enum so we need
    # to perform this workaround.
    dict_of_base_model["liquid_properties"][LIQUID_CLASS]["supportedTips"] = {
        k.name: v
        for k, v in dict_of_base_model["liquid_properties"][LIQUID_CLASS][
            "supportedTips"
        ].items()
    }
    dict_of_base_model["liquid_properties"] = {
        k.name: v for k, v in dict_of_base_model["liquid_properties"].items()
    }
    dict_of_base_model["plungerPositionsConfigurations"] = {
        k.name: v
        for k, v in dict_of_base_model["plungerPositionsConfigurations"].items()
    }
    return PipetteConfigurations.model_validate(dict_of_base_model)


def _load_available_overrides(
    pipette_serial_number: str, pipette_override_path: Path
) -> OverrideType:
    """Load the available overrides from disk."""
    pipette_override = pipette_override_path / f"{pipette_serial_number}.json"
    try:
        with open(pipette_override, "r") as f:
            return json.load(f, cls=MutableConfigurationDecoder)
    except json.JSONDecodeError as e:
        log.warning(f"pipette override for {pipette_serial_number} is corrupt: {e}")
        pipette_override.unlink()
        raise FileNotFoundError()


def _list_all_mutable_configs(
    on_disk_overrides: OverrideType, base_configurations: Dict[str, Any]
) -> OverrideType:
    """Get all available mutable configurations"""
    all_mutable_configs = set([*MUTABLE_CONFIGS_V1, *VALID_QUIRKS]) - set(["quirks"])
    on_disk_configs = set(
        [
            v.name
            for v in on_disk_overrides.values()
            if isinstance(v, MutableConfig) or isinstance(v, QuirkConfig)
        ]
    )

    missing_configurations = all_mutable_configs.difference(on_disk_configs)

    default_configurations: OverrideType = {"quirks": {}}
    for c in missing_configurations:
        if c in VALID_QUIRKS:
            if Quirks(c) in base_configurations["quirks"]:
                default_configurations["quirks"][c] = QuirkConfig.validate_and_build(  # type: ignore
                    c, True
                )
        else:
            default_configurations[c] = _find_default(c, base_configurations)
    default_configurations.update(on_disk_overrides)
    return default_configurations


def _get_default_value_for(  # noqa: C901
    config: Dict[str, Any], keypath: List[str]
) -> Any:
    def _do_get_default_value_for(
        remaining_config: Dict[Any, Any], keypath: List[Any]
    ) -> None:
        first: Any = keypath[0]
        if first in [lc.name for lc in LiquidClasses]:
            first = LiquidClasses[first]
        if len(keypath) > 1:
            rest = keypath[1:]
            if first == "##EACHTIP##":
                tip_list = list(remaining_config.keys())
                tip_list.sort(key=lambda o: o.value if isinstance(o, Enum) else o)
                return _do_get_default_value_for(remaining_config[tip_list[-1]], rest)
            elif first == "##EACHNOZZLEMAP##":
                map_list = list(remaining_config.keys())
                return _do_get_default_value_for(remaining_config[map_list[-1]], rest)
            elif first == "##EACHTIPTYPE##":
                for key in remaining_config.keys():
                    if key == "default":
                        return _do_get_default_value_for(remaining_config[key], rest)
            else:
                return _do_get_default_value_for(remaining_config[first], rest)
        else:
            if first == "##EACHTIP##":
                tip_list = list(remaining_config.keys())
                tip_list.sort(key=lambda o: o.value if isinstance(o, Enum) else o)
                return remaining_config[tip_list[-1]]
            return remaining_config[first]

    return _do_get_default_value_for(config, keypath)


def _find_default(name: str, configs: Dict[str, Any]) -> MutableConfig:
    """Find the default value from the configs and return it as a mutable config."""
    keypath = _MAP_KEY_TO_V2[name]
    nested_name = keypath[-1]

    name_to_lookup_key_map = {
        "pickUpCurrent": "current",
        "pickUpDistance": "distance",
        "pickUpSpeed": "speed",
    }
    if name in name_to_lookup_key_map.keys():
        lookup_key = name_to_lookup_key_map[name]
        min_max_dict = _MIN_MAX_LOOKUP[lookup_key]
        type_lookup = _TYPE_LOOKUP[lookup_key]
        units_lookup = _UNITS_LOOKUP[lookup_key]
    else:
        min_max_dict = _MIN_MAX_LOOKUP[nested_name]
        type_lookup = _TYPE_LOOKUP[nested_name]
        units_lookup = _UNITS_LOOKUP[nested_name]
    default_value = _get_default_value_for(configs, keypath)
    return MutableConfig(
        value=default_value,
        default=default_value,
        min=min_max_dict["min"],
        max=min_max_dict["max"],
        type=type_lookup,
        units=units_lookup,
        name=name,
    )


def known_pipettes(pipette_override_path: Path) -> List[str]:
    """List pipette IDs for which we have known overrides"""
    return [
        fi.stem
        for fi in pipette_override_path.iterdir()
        if fi.is_file() and ".json" in fi.suffixes
    ]


def _load_full_mutable_configs(
    pipette_model: PipetteModelVersionType, overrides: OverrideType
) -> OverrideType:
    base_configs = load_definition(
        pipette_model.pipette_type,
        pipette_model.pipette_channels,
        pipette_model.pipette_version,
        pipette_model.oem_type,
    )
    base_configs_dict = base_configs.model_dump(by_alias=True)
    full_mutable_configs = _list_all_mutable_configs(overrides, base_configs_dict)

    if not full_mutable_configs.get("name"):
        full_mutable_configs["name"] = str(
            convert_to_pipette_name_type(cast(PipetteName, str(pipette_model)))
        )

    if not full_mutable_configs.get("model"):
        full_mutable_configs["model"] = str(pipette_model)

    return full_mutable_configs


def list_mutable_configs(
    pipette_serial_number: str, pipette_override_path: Path
) -> OverrideType:
    """
    Returns dict of mutable configs only.
    """

    mutable_configs: OverrideType = {}

    try:
        mutable_configs = _load_available_overrides(
            pipette_serial_number, pipette_override_path
        )
    except FileNotFoundError:
        log.info(f"Pipette id {pipette_serial_number} not found")
        # This is mimicing behavior from the original file
        # which returned an empty dict if no on disk value was found.
        return mutable_configs

    serial_key_match = SERIAL_STUB_REGEX.match(pipette_serial_number)

    if serial_key_match:
        serial_key = serial_key_match.group(0)
    else:
        serial_key = ""
    pipette_model = convert_pipette_model(
        cast(PipetteModel, PIPETTE_SERIAL_MODEL_LOOKUP[serial_key])
    )
    return _load_full_mutable_configs(pipette_model, mutable_configs)


def list_mutable_configs_with_defaults(
    pipette_model: PipetteModelVersionType,
    pipette_serial_number: Optional[str],
    pipette_override_path: Path,
) -> OverrideType:
    """
    Returns dict of mutable configs only, with their defaults.
    """
    mutable_configs: OverrideType = {}
    if pipette_serial_number:
        try:
            mutable_configs = _load_available_overrides(
                pipette_serial_number, pipette_override_path
            )
        except FileNotFoundError:
            pass
    return _load_full_mutable_configs(pipette_model, mutable_configs)


def load_with_mutable_configurations(
    pipette_model: PipetteModelVersionType,
    pipette_override_path: Path,
    pipette_serial_number: Optional[str] = None,
) -> PipetteConfigurations:
    """
    Load pipette config data with any overrides available.

    This function reads from disk each time, so changes to the overrides
    will be picked up in subsequent calls.

    :param str pipette_model: The pipette model name (i.e. "p10_single_v1.3")
                              for which to load configuration
    :param pipette_override_path: The path to the on-disk file which has the config overrides.
    :param pipette_serial_number: An (optional) unique ID for the pipette to locate
                       config overrides. If the ID is not specified, the system
                       assumes this is a simulated pipette and does not
                       save settings. If the ID is specified but no overrides
                       corresponding to the ID are found, the system creates a
                       new overrides file for it.
    :type pipette_serial_number: str or None
    :raises KeyError: if ``pipette_model`` is not in the top-level keys of
                      the pipetteModelSpecs.json file (and therefore not in
                      :py:attr:`configs`)

    :returns PipetteConfig: The configuration, loaded and checked
    """
    base_configurations = load_definition(
        pipette_model.pipette_type,
        pipette_model.pipette_channels,
        pipette_model.pipette_version,
        pipette_model.oem_type,
    )
    # Load overrides if we have a pipette id
    if pipette_serial_number:
        try:
            override = _load_available_overrides(
                pipette_serial_number, pipette_override_path
            )
        except FileNotFoundError:
            pass
        else:
            try:
                base_configurations = _migrate_to_v2_configurations(
                    base_configurations, override
                )
            except BaseException:
                log.exception(
                    "Failed to migrate mutable configurations. Please report this as it is a bug."
                )

    # the ulPerMm functions are structured in pipetteModelSpecs.json as
    # a list sorted from oldest to newest. That means the latest functions
    # are always the last element and, as of right now, the older ones are
    # the first element (for models that only have one function, the first
    # and last elements are the same, which is fine). If we add more in the
    # future, we’ll have to change this code to select items more
    # intelligently
    # TODO (lc 06/27) Handle versioned ul per mm

    return base_configurations


def _add_new_overrides_to_existing(  # noqa: C901
    base_configs_dict: Dict[str, Any],
    existing_overrides: OverrideType,
    overrides: TypeOverrides,
) -> OverrideType:
    """Helper function to add new overrides to the existing ones"""
    # FIXME remove the validation here for the file save and rely
    # on the validation in the robot server. We unfortunately have
    # to keep the validation here until we decide to fully wipe/migrate
    # files saved on disk because some of them have unexpected
    # data entries.
    if not existing_overrides.get("quirks"):
        existing_overrides["quirks"] = {}
    for key, value in overrides.items():
        # If an existing override is saved as null from endpoint, remove from
        # overrides file
        if value is None and existing_overrides.get(key):
            del existing_overrides[key]
        elif isinstance(value, bool):
            if key in VALID_QUIRKS:
                existing_overrides["quirks"][key] = QuirkConfig.validate_and_build(  # type: ignore
                    key, value
                )
            elif key not in MUTABLE_CONFIGS_V1:
                # Unfortunately, some of the files got corrupted,
                # so we have to check that the key doesn't exist
                # in mutable configs before throwing an error.
                raise ValueError(
                    f"{value} is invalid for {key} or {key} is not a supported quirk."
                )
            elif existing_overrides.get(key):
                del existing_overrides[key]
        elif value:
            if existing_overrides.get(key):
                existing_overrides[key].validate_and_add(value)  # type: ignore
            else:
                new_mutable_config = _find_default(key, base_configs_dict)
                new_mutable_config.validate_and_add(value)
                existing_overrides[key] = new_mutable_config
    return existing_overrides


def save_overrides(
    pipette_serial_number: str, overrides: TypeOverrides, pipette_override_path: Path
) -> None:
    """
    Save overrides for the pipette.

    :param pipette_id: The pipette id
    :param overrides: The incoming values
    :return: None
    """
    # need to load defaults
    serial_key_match = SERIAL_STUB_REGEX.match(pipette_serial_number)
    if serial_key_match:
        serial_key = serial_key_match.group(0)
    else:
        serial_key = ""
    pipette_model = convert_pipette_model(
        cast(PipetteModel, PIPETTE_SERIAL_MODEL_LOOKUP[serial_key])
    )

    base_configs = load_definition(
        pipette_model.pipette_type,
        pipette_model.pipette_channels,
        pipette_model.pipette_version,
        pipette_model.oem_type,
    )
    base_configs_dict = base_configs.model_dump(by_alias=True)
    try:
        existing_overrides = _load_available_overrides(
            pipette_serial_number, pipette_override_path
        )
    except FileNotFoundError:
        existing_overrides = {"quirks": {}}

    updated_overrides = _add_new_overrides_to_existing(
        base_configs_dict, existing_overrides, overrides
    )

    if not updated_overrides.get("model"):
        updated_overrides["model"] = str(pipette_model)

    with open(pipette_override_path / f"{pipette_serial_number}.json", "w") as file:
        json.dump(updated_overrides, file, cls=MutableConfigurationEncoder)



from typing import Optional, List, Dict, Any, Literal
from enum import Enum
from pydantic import BaseModel


class Dimensions(BaseModel):
    xDimension: float
    yDimension: float
    zDimension: float


class GroupMetadata(BaseModel):
    pass


class Group(BaseModel):
    metadata: GroupMetadata
    wells: List[str]


class Shape(Enum):
    circular = "circular"
    rectangular = "rectangular"


class WellDefinition(BaseModel):
    depth: float
    totalLiquidVolume: int
    shape: Shape
    x: float
    y: float
    z: float
    diameter: Optional[float] = None
    yDimension: Optional[float] = None
    xDimension: Optional[float] = None


class Metadata(BaseModel):
    protocolName: Optional[str] = None
    author: Optional[str] = None
    description: Optional[str] = None
    created: Optional[int] = None
    lastModified: Optional[int] = None
    category: Optional[str] = None
    subcategory: Optional[str] = None
    tags: Optional[List[str]] = None


class Module(BaseModel):
    model: str


class Pipette(BaseModel):
    name: str


class Robot(BaseModel):
    model: Literal["OT-2 Standard", "OT-3 Standard"]
    deckId: str


class DesignerApplication(BaseModel):
    name: Optional[str] = None
    version: Optional[str] = None
    data: Optional[Dict[str, Any]] = None


class CommandAnnotation(BaseModel):
    commandIds: List[str]
    annotationType: str


class OffsetVector(BaseModel):
    x: Optional[float] = None
    y: Optional[float] = None
    z: Optional[float] = None


class Location(BaseModel):
    slotName: Optional[str] = None
    moduleId: Optional[str] = None
    labwareId: Optional[str] = None
    addressableAreaName: Optional[str] = None


class ProfileStep(BaseModel):
    celsius: float
    holdSeconds: float


class WellLocation(BaseModel):
    origin: Optional[str] = None
    offset: Optional[OffsetVector] = None


class Liquid(BaseModel):
    displayName: str
    description: str
    displayColor: Optional[str] = None


class Labware(BaseModel):
    displayName: Optional[str] = None
    definitionId: str


class NozzleConfigurationParams(BaseModel):
    style: str
    primaryNozzle: Optional[str] = None
    frontRightNozzle: Optional[str] = None



from opentrons_shared_data.labware import list_definitions


def test_list_definitions() -> None:
    """A very basic test on list_definitions to make sure it's returning _something_,
    i.e. it doesn't use a wrong hard-coded path or something like that.
    """
    assert len(list_definitions()) > 0



"""Gripper constants and default values."""

LABWARE_GRIP_FORCE = 15.0  # Newtons
GRIPPER_DECK_DROP_OFFSET = [-0.5, -0.5, -1.5]
GRIPPER_MODULE_DROP_OFFSET = [0.0, 0.0, 1.5]
GRIPPER_PADDLE_WIDTH = 36  # mm



from typing import Dict, Generator, List, Set, Tuple

import pytest

from opentrons_shared_data.deck import (
    list_names as list_deck_definition_names,
    load as load_deck_definition,
)
from opentrons_shared_data.deck.types import (
    AddressableArea,
    Cutout,
    CutoutFixture,
    DeckDefinitionV3,
    DeckDefinitionV4,
)


def as_tuple(list: List[float]) -> Tuple[float, float, float]:
    """Convert an [x,y,z] list from the definitions to an (x,y,z) tuple, for hashability."""
    [x, y, z] = list
    return (x, y, z)


def get_v3_slot_positions(
    definition: DeckDefinitionV3,
) -> Set[Tuple[str, Tuple[float, float, float]]]:
    """Return all the slot positions defined by the deck definition, as (slot_id, [x,y,z]) tuples."""
    return set(
        (ordered_slot["id"], as_tuple(ordered_slot["position"]))
        for ordered_slot in definition["locations"]["orderedSlots"]
    )


def get_v4_stacks(
    definition: DeckDefinitionV4,
) -> Generator[Tuple[Cutout, CutoutFixture, AddressableArea], None, None]:
    """Yield all the (cutout, cutoutFixture, addressableArea) combinations that the def allows."""
    cutout_fixtures = definition["cutoutFixtures"]
    cutouts_by_id: Dict[str, Cutout] = {
        cutout["id"]: cutout for cutout in definition["locations"]["cutouts"]
    }
    addressable_areas_by_id: Dict[str, AddressableArea] = {
        addressable_area["id"]: addressable_area
        for addressable_area in definition["locations"]["addressableAreas"]
    }

    for cutout_fixture in cutout_fixtures:
        for cutout_id, addressable_area_ids in cutout_fixture[
            "providesAddressableAreas"
        ].items():
            for addressable_area_id in addressable_area_ids:
                cutout = cutouts_by_id[cutout_id]
                addressable_area = addressable_areas_by_id[addressable_area_id]
                yield cutout, cutout_fixture, addressable_area


def compute_v4_position(
    cutout: Cutout, addressable_area: AddressableArea
) -> List[float]:
    return [
        a + b
        for a, b in zip(cutout["position"], addressable_area["offsetFromCutoutFixture"])
    ]


def get_v4_slot_positions(
    definition: DeckDefinitionV4,
) -> Set[Tuple[str, Tuple[float, float, float]]]:
    """Return all the slot positions defined by the deck definition, as (addressable_area_id, [x,y,z]) tuples.

    It's important that this returns a set of (addressable_area_id, [x,y,z]) tuples instead of a
    dict {addressable_area_id: [x,y,z]}. Deck schema v4 theoretically allows a single addressable
    area ID to be associated with multiple positions. If that happens, we don't want to accidentally
    overwrite any.
    """
    stacks_with_slots = (
        (cutout, cutout_fixture, addressable_area)
        for (cutout, cutout_fixture, addressable_area) in get_v4_stacks(definition)
        if addressable_area["areaType"] == "slot"
    )

    slot_positions = (
        (
            addressable_area["id"],
            as_tuple(compute_v4_position(cutout, addressable_area)),
        )
        for (cutout, cutout_fixture, addressable_area) in stacks_with_slots
    )

    return set(slot_positions)


@pytest.mark.parametrize("definition_name", list_deck_definition_names(version=4))
def test_v3_and_v4_positional_equivalence(definition_name: str) -> None:
    deck_v3 = load_deck_definition(name=definition_name, version=3)
    deck_v4 = load_deck_definition(name=definition_name, version=4)

    v3_slot_positions = get_v3_slot_positions(deck_v3)
    v4_slot_positions = get_v4_slot_positions(deck_v4)

    assert v3_slot_positions == v4_slot_positions



"""Gripper configurations."""

from typing_extensions import Annotated, Literal
from typing import List, Dict, Tuple, Any, NewType
from pydantic import ConfigDict, BaseModel, Field
from enum import Enum


def _snake_to_camel_case(snake: str) -> str:
    """Turns snake_case to camelCase."""
    return "".join(
        [s.capitalize() if i > 0 else s.lower() for i, s in enumerate(snake.split("_"))]
    )


GripperModelStr = NewType("GripperModelStr", str)


# TODO (spp, 2023-01-31): figure out if we want to keep this a string enum or revert to
#  a regular enum with custom stringification
class GripperModel(str, Enum):
    """Gripper models."""

    v1 = "gripperV1"
    v1_1 = "gripperV1.1"
    v1_2 = "gripperV1.2"
    v1_3 = "gripperV1.3"

    def __str__(self) -> str:
        """Model name."""
        enum_to_str = {
            self.__class__.v1: "gripperV1",
            self.__class__.v1_1: "gripperV1.1",
            self.__class__.v1_2: "gripperV1.2",
            self.__class__.v1_3: "gripperV1.3",
        }
        return enum_to_str[self]


GripperSchemaVersion = Literal[1]

GripperSchema = Dict[str, Any]


_StrictNonNegativeInt = Annotated[int, Field(strict=True, ge=0)]
_StrictNonNegativeFloat = Annotated[float, Field(strict=True, ge=0.0)]


PolynomialTerm = Tuple[_StrictNonNegativeInt, float]
_Polynomial = Annotated[List[PolynomialTerm], Field(min_length=1)]


class GripperBaseModel(BaseModel):
    """Gripper base model."""

    model_config = ConfigDict(
        alias_generator=_snake_to_camel_case, populate_by_name=True
    )


Offset = Tuple[float, float, float]


class Geometry(GripperBaseModel):
    """Gripper geometry definition."""

    base_offset_from_mount: Offset
    jaw_center_offset_from_base: Offset
    pin_one_offset_from_base: Offset
    pin_two_offset_from_base: Offset
    jaw_width: Dict[str, float]
    max_allowed_grip_error: _StrictNonNegativeFloat


class GripForceProfile(GripperBaseModel):
    """Gripper force profile."""

    polynomial: _Polynomial = Field(
        ...,
        description="Polynomial function to convert a grip force in Newton to the jaw motor duty cycle value, which will be read by the gripper firmware.",
    )
    default_grip_force: _StrictNonNegativeFloat
    default_idle_force: _StrictNonNegativeFloat
    default_home_force: _StrictNonNegativeFloat
    min: _StrictNonNegativeFloat
    max: _StrictNonNegativeFloat


class GripperDefinition(GripperBaseModel):
    """Gripper definition."""

    schema_version: GripperSchemaVersion = Field(
        ..., description="Which schema version a gripper is using"
    )
    display_name: str = Field(..., description="Gripper display name.")
    model: GripperModel
    geometry: Geometry
    grip_force_profile: GripForceProfile



import typing
import logging
import os
from pathlib import Path
from functools import lru_cache

log = logging.getLogger(__name__)

ENV_SHARED_DATA_PATH = "OT_SHARED_DATA_PATH"


class SharedDataMissingError(IOError):
    pass


@lru_cache(maxsize=1)
def get_shared_data_root() -> Path:
    """
    Get the root directory of the shared data.

    Steps (first to succeed wins):
    1) Use environment variable in OT_SHARED_DATA_PATH
    2) Look in "shared_data" in the root of the installed package
    3) Look for "shared-data" in parent directories.
    4) Raise exception
    """
    # Check environment variable
    override = os.environ.get(ENV_SHARED_DATA_PATH)
    if override is not None:
        log.info("Using override for shared data path: %s", override)
        return Path(override)

    # Check contents of package
    module_path = Path(__file__).parent
    module_data = module_path / "data"
    if module_data.exists():
        log.info(f"Using packaged shared data path: {str(module_data)}")
        return module_data

    # We are likely to be running locally and will find shared-data in repo
    for parent in module_path.parents:
        p = parent / "shared-data"
        if p.exists():
            log.info("Using shared data in path: %s", p)
            return p

    raise SharedDataMissingError()


def load_shared_data(path: typing.Union[str, Path]) -> bytes:
    """
    Load file from shared data directory.

    path is relative to the root of all shared data (ie. no "shared-data")
    """
    with open(get_shared_data_root() / path, "rb") as f:
        return f.read()



"""Type definitions for liquid classes."""
from typing import Sequence, Tuple, TypedDict
from typing_extensions import NotRequired
from .liquid_class_definition import (
    POSITION_REFERENCE_VALUE_TYPE,
    BLOWOUT_LOCATION_VALUE_TYPE,
)


class Offset(TypedDict):
    """A dict representing an offset."""

    x: float
    y: float
    z: float


class TipPositionDict(TypedDict):
    """A dict representing a tip position."""

    position_reference: POSITION_REFERENCE_VALUE_TYPE
    offset: Offset


class DelayPropertiesDict(TypedDict):
    """A dict representing a delay."""

    enabled: bool
    duration: NotRequired[float]


class TouchTipPropertiesDict(TypedDict):
    """A dict representing touch tip properties."""

    enabled: bool
    z_offset: NotRequired[float]
    mm_from_edge: NotRequired[float]
    speed: NotRequired[float]


class MixPropertiesDict(TypedDict):
    """A dict representing mix properties."""

    enabled: bool
    repetitions: NotRequired[int]
    volume: NotRequired[float]


class BlowoutPropertiesDict(TypedDict):
    """A dict representing blowout properties."""

    enabled: bool
    location: NotRequired[BLOWOUT_LOCATION_VALUE_TYPE]
    flow_rate: NotRequired[float]


class SubmergeDict(TypedDict):
    """A dict representing submerge properties."""

    start_position: TipPositionDict
    speed: float
    delay: DelayPropertiesDict


class RetractAspirateDict(TypedDict):
    """A dict representing retract aspirate properties."""

    end_position: TipPositionDict
    speed: float
    delay: DelayPropertiesDict
    air_gap_by_volume: Sequence[Tuple[float, float]]
    touch_tip: TouchTipPropertiesDict


class RetractDispenseDict(TypedDict):
    """A dict representing retract dispense properties."""

    end_position: TipPositionDict
    speed: float
    delay: DelayPropertiesDict
    air_gap_by_volume: Sequence[Tuple[float, float]]
    touch_tip: TouchTipPropertiesDict
    blowout: BlowoutPropertiesDict


class AspiratePropertiesDict(TypedDict):
    """A dict representing aspirate properties."""

    submerge: SubmergeDict
    flow_rate_by_volume: Sequence[Tuple[float, float]]
    correction_by_volume: Sequence[Tuple[float, float]]
    delay: DelayPropertiesDict
    aspirate_position: TipPositionDict
    retract: RetractAspirateDict
    pre_wet: bool
    mix: MixPropertiesDict


class SingleDispensePropertiesDict(TypedDict):
    """A dict representing single dispense properties."""

    submerge: SubmergeDict
    flow_rate_by_volume: Sequence[Tuple[float, float]]
    correction_by_volume: Sequence[Tuple[float, float]]
    delay: DelayPropertiesDict
    dispense_position: TipPositionDict
    retract: RetractDispenseDict
    push_out_by_volume: Sequence[Tuple[float, float]]
    mix: MixPropertiesDict


class MultiDispensePropertiesDict(TypedDict):
    """A dict representing multi dispense properties."""

    submerge: SubmergeDict
    flow_rate_by_volume: Sequence[Tuple[float, float]]
    correction_by_volume: Sequence[Tuple[float, float]]
    delay: DelayPropertiesDict
    dispense_position: TipPositionDict
    retract: RetractDispenseDict
    conditioning_by_volume: Sequence[Tuple[float, float]]
    disposal_by_volume: Sequence[Tuple[float, float]]


class TransferPropertiesDict(TypedDict):
    """A dict representing transfer properties for a specific pipette and tiprack."""

    aspirate: AspiratePropertiesDict
    dispense: SingleDispensePropertiesDict
    multi_dispense: NotRequired[MultiDispensePropertiesDict]



import os
import pytest
from typing import Iterator
from opentrons_shared_data import get_shared_data_root
from opentrons_shared_data.pipette.pipette_load_name_conversions import (
    convert_pipette_model,
)
from opentrons_shared_data.pipette.load_data import load_definition
from opentrons_shared_data.pipette.ul_per_mm import piecewise_volume_conversion

from opentrons_shared_data.pipette.types import PipetteModel
from opentrons_shared_data.pipette.pipette_definition import (
    ulPerMMDefinition,
)


DEFAULT_MAX_SPEED_HIGH_THROUGHPUT_OT3_AXIS_KIND_P = 15
DEFAULT_MAX_SPEED_LOW_THROUGHPUT_OT3_AXIS_KIND_P = 70
B_MAX_SPEED = 40


def _get_plunger_max_speed(pipette_model: PipetteModel) -> float:
    if "v2" in pipette_model:
        return B_MAX_SPEED
    else:
        if "96" in pipette_model:
            return DEFAULT_MAX_SPEED_HIGH_THROUGHPUT_OT3_AXIS_KIND_P
        else:
            return DEFAULT_MAX_SPEED_LOW_THROUGHPUT_OT3_AXIS_KIND_P


def _get_max_flow_rate_at_volume(
    ul_per_mm_definition: ulPerMMDefinition,
    pipette_model: PipetteModel,
    volume: float,
) -> float:
    max_speed = _get_plunger_max_speed(pipette_model)
    map = list(ul_per_mm_definition.default.values())[-1]
    ul_per_mm = piecewise_volume_conversion(volume, map)
    return round(ul_per_mm * max_speed, 1)


def get_all_pipette_models() -> Iterator[PipetteModel]:
    paths_to_validate = (
        get_shared_data_root() / "pipette" / "definitions" / "2" / "liquid"
    )

    _channel_model_str = {
        "single_channel": "single",
        "ninety_six_channel": "96",
        "eight_channel": "multi",
        "eight_channel_em": "multi_em",
    }
    assert os.listdir(paths_to_validate), "You have a path wrong"
    for channel_dir in os.listdir(paths_to_validate):
        for model_dir in os.listdir(paths_to_validate / channel_dir):
            for liquid_file in os.listdir(paths_to_validate / channel_dir / model_dir):
                for version_file in os.listdir(
                    paths_to_validate / channel_dir / model_dir / liquid_file
                ):
                    version_list = version_file.split(".json")[0].split("_")
                    built_model: PipetteModel = PipetteModel(
                        f"{model_dir}_{_channel_model_str[channel_dir]}_v{version_list[0]}.{version_list[1]}"
                    )
                    if version_list[0] != "1" and version_list[1] != "0":
                        yield built_model


@pytest.mark.parametrize("pipette", list(get_all_pipette_models()))
@pytest.mark.parametrize("action", ["aspirate", "dispense"])
def test_max_flow_rates_per_volume(pipette: PipetteModel, action: str) -> None:
    """Verify the max flow rate values for each pipette's supported tip is in range"""
    pipette_model_version = convert_pipette_model(pipette)
    definition = load_definition(
        pipette_model_version.pipette_type,
        pipette_model_version.pipette_channels,
        pipette_model_version.pipette_version,
        pipette_model_version.oem_type,
    )

    pipette_model_version_str = f"{pipette_model_version}"

    for liquid_name, liquid_properties in definition.liquid_properties.items():
        for tip_type, supported_tip in liquid_properties.supported_tips.items():

            """TODO: the following models do not pass the asserts since the uiMaxFlowRate was raised
            to match the default blowout and dispense flowRates. uiMaxFlowRate will be reevaluated
            in the future."""
            if not (
                (
                    pipette_model_version_str
                    in {
                        "p50_single_v3.4",
                        "p50_single_v3.5",
                        "p50_single_v3.6",
                        "p50_multi_v3.5",
                        "p50_multi_v3.4",
                    }
                    and liquid_properties.min_volume == 5.0
                )
                or (
                    pipette_model_version_str
                    in {"p200_96_v3.0", "p200_96_v3.1", "p200_96_v3.2"}
                    and liquid_properties.min_volume == 0.5
                )
            ):
                assert supported_tip.ui_max_flow_rate <= _get_max_flow_rate_at_volume(
                    supported_tip.aspirate, pipette, liquid_properties.min_volume
                )
                assert supported_tip.ui_max_flow_rate <= _get_max_flow_rate_at_volume(
                    supported_tip.dispense, pipette, liquid_properties.min_volume
                )



import pytest
from pydantic import ValidationError
import json
from opentrons_shared_data import gripper, load_shared_data
from opentrons_shared_data.gripper import (
    GripperDefinition,
    GripperModel,
)

GRIPPER_DEF = json.loads(load_shared_data("gripper/definitions/1/gripperV1.json"))


def test_gripper_definition() -> None:
    gripper_def = gripper.load_definition(GripperModel.v1)
    assert isinstance(gripper_def, GripperDefinition)
    assert gripper_def.schema_version == 1


def test_gripper_definition_type() -> None:
    assert GripperDefinition.model_validate(GRIPPER_DEF)

    # missing key
    del GRIPPER_DEF["gripForceProfile"]
    with pytest.raises(ValidationError):
        assert GripperDefinition.model_validate(GRIPPER_DEF)

    # no missing key but with incorrect value
    GRIPPER_DEF["geometry"]["gripForceProfile"] = {"min": 1.0, "max": "0.0"}
    with pytest.raises(ValidationError):
        assert GripperDefinition.model_validate(GRIPPER_DEF)



import json
import pytest
from pathlib import Path

from opentrons_shared_data import load_shared_data
from opentrons_shared_data.protocol.models import protocol_schema_v8

from . import list_fixtures


@pytest.mark.parametrize("defpath", list_fixtures(8))
def test_v8_types(defpath: Path) -> None:
    def_data = load_shared_data(defpath)
    def_model = protocol_schema_v8.ProtocolSchemaV8.model_validate_json(def_data)
    def_dict_from_model = def_model.model_dump(by_alias=True, exclude_unset=True)
    expected_def_dict = json.loads(def_data)
    assert def_dict_from_model == expected_def_dict



"""opentrons_shared_data.robot: Submodule for handling robot definition data."""
from pathlib import Path
from typing import cast
from typing_extensions import Final

import json

from .. import get_shared_data_root

from .types import RobotDefinition, RobotType

DEFAULT_ROBOT_DEFINITION_VERSION: Final = 1


def load(
    robot_type: RobotType, version: int = DEFAULT_ROBOT_DEFINITION_VERSION
) -> RobotDefinition:
    """Load the definition for the specified robot id."""
    for fi in Path(
        get_shared_data_root() / "robot" / "definitions" / f"{version}"
    ).iterdir():
        defn = json.load(fi.open("r"))
        if defn["robotType"] == robot_type:
            return cast(RobotDefinition, defn)
    raise KeyError(robot_type)


def user_facing_robot_type(robot_type: RobotType, include_article: bool = False) -> str:
    """Appropriately formatted robot type string for use in user-facing messages."""
    if robot_type == "OT-2 Standard":
        return "an OT-2" if include_article else "OT-2"
    elif robot_type == "OT-3 Standard":
        return "a Flex" if include_article else "Flex"



import re
from typing import List, Dict, Tuple, Optional, Annotated, Literal, TypeVar
from pydantic import (
    field_validator,
    BaseModel,
    Field,
    BeforeValidator,
    PlainSerializer,
)
from dataclasses import dataclass

from . import types as pip_types, types

# The highest and lowest existing overlap version values.
TIP_OVERLAP_VERSION_MINIMUM = 0
TIP_OVERLAP_VERSION_MAXIMUM = 3

PLUNGER_CURRENT_MINIMUM = 0.1
PLUNGER_CURRENT_MAXIMUM = 1.5

NOZZLE_MAP_NAMES = re.compile(r"(?P<row>[A-Z]+)(?P<column>[0-9]+)")
COLUMN_NAMES = re.compile(r"[0-9]+")
ROW_NAMES = re.compile(r"[A-Z]+")
OT_TIPRACK_NAMES = re.compile(r"opentrons/[a-z0-9._]+/[0-9]")


def validate_opentrons_tiprack(v: str) -> str:
    if not OT_TIPRACK_NAMES.match(v):
        raise ValueError("{v} is not a valid tiprack name.")
    return v


EnumType = TypeVar("EnumType")
EnumSerializer = Annotated[EnumType, PlainSerializer(lambda v: v.value)]


# TODO (lc 12-5-2022) Ideally we can deprecate this
# at somepoint once we load pipettes by channels and type
@dataclass
class PipetteNameType:
    pipette_type: pip_types.PipetteModelType
    pipette_channels: pip_types.PipetteChannelType
    pipette_generation: pip_types.PipetteGenerationType
    oem_type: pip_types.PipetteOEMType

    def __repr__(self) -> str:
        oem_name = (
            f"_{self.oem_type.value}"
            if self.oem_type != pip_types.PipetteOEMType.OT
            else ""
        )
        base_name = f"{self.pipette_type.name}_{str(self.pipette_channels)}{oem_name}"
        if self.pipette_generation == pip_types.PipetteGenerationType.GEN1:
            return base_name
        elif self.pipette_channels == pip_types.PipetteChannelType.NINETY_SIX_CHANNEL:
            return base_name
        else:
            return f"{base_name}_{self.pipette_generation.name.lower()}"

    def get_version(self) -> pip_types.PipetteVersionType:
        if self.pipette_generation == pip_types.PipetteGenerationType.FLEX:
            return pip_types.PipetteVersionType(3, 0)
        elif self.pipette_generation == pip_types.PipetteGenerationType.GEN2:
            return pip_types.PipetteVersionType(2, 0)
        else:
            return pip_types.PipetteVersionType(1, 0)


@dataclass
class PipetteModelVersionType:
    pipette_type: pip_types.PipetteModelType
    pipette_channels: pip_types.PipetteChannelType
    pipette_version: pip_types.PipetteVersionType
    oem_type: pip_types.PipetteOEMType

    def __repr__(self) -> str:
        oem_name = (
            f"_{self.oem_type.value}"
            if self.oem_type != pip_types.PipetteOEMType.OT
            else ""
        )
        base_name = f"{self.pipette_type.name}_{str(self.pipette_channels)}{oem_name}"

        return f"{base_name}_v{self.pipette_version}"


class FlowRateDefinition(BaseModel):
    default: float = Field(..., description="Highest API level default fallback.")
    values_by_api_level: Dict[str, float] = Field(
        ..., description="flow rates keyed by API level.", alias="valuesByApiLevel"
    )


PipetteFunctionKeyType = Literal["1", "2", "3"]
ulPerMMType = Dict[PipetteFunctionKeyType, List[Tuple[float, float, float]]]


class ulPerMMDefinition(BaseModel):
    default: ulPerMMType


class SupportedTipsDefinition(BaseModel):
    """Tip parameters available for every tip size."""

    default_aspirate_flowrate: FlowRateDefinition = Field(
        ...,
        description="The flowrate used in aspirations by default. For lowVolumeDefault only, the flowrate matches uiMaxFlowRate for ui purposes, it does not change physical behavior.",
        alias="defaultAspirateFlowRate",
    )
    default_dispense_flowrate: FlowRateDefinition = Field(
        ...,
        description="The flowrate used in dispenses by default. For lowVolumeDefault only, the flowrate matches uiMaxFlowRate for ui purposes, it does not change physical behavior.",
        alias="defaultDispenseFlowRate",
    )
    default_blowout_flowrate: FlowRateDefinition = Field(
        ...,
        description="The flowrate used in blowouts by default. For lowVolumeDefault only, the flowrate matches uiMaxFlowRate for ui purposes, it does not change physical behavior.",
        alias="defaultBlowOutFlowRate",
    )
    default_flow_acceleration: float = Field(
        float("inf"),  # no default works for all pipettes
        description="The acceleration used during aspirate/dispense/blowout in ul/s^2.",
        alias="defaultFlowAcceleration",
    )
    default_tip_length: float = Field(
        ...,
        description="The default tip length associated with this tip type.",
        alias="defaultTipLength",
    )
    default_return_tip_height: float = Field(
        default=0.5,
        description="The height to return a tip to its tiprack.",
        alias="defaultReturnTipHeight",
    )
    aspirate: ulPerMMDefinition = Field(
        ..., description="The default pipetting functions list for aspirate."
    )
    dispense: ulPerMMDefinition = Field(
        ..., description="The default pipetting functions list for dispensing."
    )
    default_push_out_volume: float = Field(
        ...,
        description="The default volume for a push-out during dispense.",
        alias="defaultPushOutVolume",
    )
    ui_max_flow_rate: float = Field(
        float(
            "inf"
        ),  # some pipettes (GEN1, unreleased prototype models) don't have a max flow rate
        description="The lowest volume max flow rate for a pipette's given supported tip, minus 2 percent for safety.",
        alias="uiMaxFlowRate",
    )


class MotorConfigurations(BaseModel):
    idle: float = Field(
        ..., description="The plunger motor current to use during idle states."
    )
    run: float = Field(
        ..., description="The plunger motor current to use during active states."
    )


class PlungerPositions(BaseModel):
    top: float = Field(
        ...,
        description="The plunger position that describes max available volume of a pipette in mm.",
    )
    bottom: float = Field(
        ...,
        description="The plunger position that describes min available volume of a pipette in mm.",
    )
    blow_out: float = Field(
        ...,
        description="The plunger position past 0 volume to blow out liquid.",
        alias="blowout",
    )
    drop_tip: float = Field(
        ..., description="The plunger position used to drop tips.", alias="drop"
    )


class PlungerHomingConfigurations(BaseModel):
    current: float = Field(
        default=0.0,
        description="The current to move the plunger axis for homing.",
    )
    speed: float = Field(
        ...,
        description="The speed to move the plunger axis for homing.",
    )


class ValidNozzleMaps(BaseModel):
    maps: Dict[str, List[str]] = Field(
        ...,
        description="Dictionary of predetermined nozzle maps for partial tip configurations.",
    )


class PressAndCamConfigurationValues(BaseModel):
    speed: float = Field(
        ...,
        description="The speed to move the Z axis for each force pickup of a given tip configuration for a given tip type.",
    )
    distance: float = Field(
        ...,
        description="The starting distance to begin a pick up tip from, based on tip configuration and tip type.",
    )
    current: float = Field(
        ...,
        description="The current used by a given tip configuration by tip type.",
    )
    versioned_tip_overlap_dictionary: Dict[str, Dict[str, float]] = Field(
        ...,
        description="Versioned default tip overlap dictionaries associated with this tip type by configuration.",
        alias="tipOverlaps",
    )


class PressFitPickUpTipConfiguration(BaseModel):
    presses: int = Field(
        ...,
        description="The number of times to force pickup (incrementally more each time by increment)",
    )
    increment: float = Field(
        ...,
        description="The increment to move the pipette down on each force tip pickup press",
    )
    configuration_by_nozzle_map: Dict[
        str, Dict[str, PressAndCamConfigurationValues]
    ] = Field(
        ...,
        description="The speed, distance, current and tip overlap configurations for a given pipette configuration. Double dictionary is keyed by Valid Nozzle Map Key and Tip Type.",
        alias="configurationsByNozzleMap",
    )


class CamActionPickUpTipConfiguration(BaseModel):
    prep_move_distance: float = Field(
        ..., description="How far to move the cams to engage the rack"
    )
    prep_move_speed: float = Field(
        ..., description="How fast to move the cams when moving to the rack"
    )
    connect_tiprack_distance_mm: float = Field(
        description="The distance to move the head down to connect with the tiprack before clamping.",
        alias="connectTiprackDistanceMM",
    )
    configuration_by_nozzle_map: Dict[
        str, Dict[str, PressAndCamConfigurationValues]
    ] = Field(
        ...,
        description="The speed, distance, current and overlap configurations for a given partial tip configuration by tip type.",
        alias="configurationsByNozzleMap",
    )


class PlungerEjectDropTipConfiguration(BaseModel):
    current: float = Field(
        ..., description="The current to use on the plunger motor when dropping a tip"
    )
    speed: float = Field(
        ..., description="How fast to move the plunger motor when dropping a tip"
    )


class CamActionDropTipConfiguration(BaseModel):
    current: float = Field(
        ..., description="The current to use on the cam motors when dropping tips"
    )
    distance: float = Field(
        ..., description="The distance to move the cams when dropping tips"
    )
    speed: float = Field(
        ..., description="How fast to move the cams when dropping tips"
    )
    prep_move_distance: float = Field(
        ..., description="How far to move the cams after disengaging"
    )
    prep_move_speed: float = Field(
        ..., description="How fast to move the cams after disengaging"
    )


class DropTipConfigurations(BaseModel):
    plunger_eject: Optional[PlungerEjectDropTipConfiguration] = Field(
        None,
        description="Configuration for tip drop via plunger eject",
        alias="plungerEject",
    )
    cam_action: Optional[CamActionDropTipConfiguration] = Field(
        None, description="Configuration for tip drop via cam action", alias="camAction"
    )


class PickUpTipConfigurations(BaseModel):
    press_fit: PressFitPickUpTipConfiguration = Field(
        description="Configuration for tip pickup via press fit", alias="pressFit"
    )
    cam_action: Optional[CamActionPickUpTipConfiguration] = Field(
        default=None,
        description="Configuration for tip pickup via cam action",
        alias="camAction",
    )


class AvailableSensorDefinition(BaseModel):
    """The number and type of sensors available in the pipette."""

    sensors: List[str] = Field(..., description="")


class PartialTipDefinition(BaseModel):
    partial_tip_supported: bool = Field(
        ...,
        description="Whether partial tip pick up is supported.",
        alias="partialTipSupported",
    )
    available_configurations: Optional[List[int]] = Field(
        default=None,
        description="A list of the types of partial tip configurations supported, listed by channel ints",
        alias="availableConfigurations",
    )


class PipettePhysicalPropertiesDefinition(BaseModel):
    """The physical properties definition of a pipette."""

    display_name: str = Field(
        ...,
        description="The display or full product name of the pipette.",
        alias="displayName",
    )
    pipette_backcompat_names: List[types.PipetteName] = Field(
        ...,
        description="A list of pipette names that are compatible with this pipette.",
        alias="backCompatNames",
    )
    pipette_type: EnumSerializer[pip_types.PipetteModelType] = Field(
        ...,
        description="The pipette model type (related to number of channels).",
        alias="model",
    )
    display_category: EnumSerializer[pip_types.PipetteGenerationType] = Field(
        ..., description="The product model of the pipette.", alias="displayCategory"
    )
    pick_up_tip_configurations: PickUpTipConfigurations = Field(
        ..., alias="pickUpTipConfigurations"
    )
    drop_tip_configurations: DropTipConfigurations = Field(
        ..., alias="dropTipConfigurations"
    )
    plunger_homing_configurations: PlungerHomingConfigurations = Field(
        ..., alias="plungerHomingConfigurations"
    )
    plunger_motor_configurations: MotorConfigurations = Field(
        ..., alias="plungerMotorConfigurations"
    )
    plunger_positions_configurations: Dict[
        pip_types.LiquidClasses, PlungerPositions
    ] = Field(..., alias="plungerPositionsConfigurations")
    available_sensors: AvailableSensorDefinition = Field(..., alias="availableSensors")
    partial_tip_configurations: PartialTipDefinition = Field(
        ..., alias="partialTipConfigurations"
    )
    channels: EnumSerializer[pip_types.PipetteChannelType] = Field(
        ..., description="The maximum number of channels on the pipette."
    )
    shaft_diameter: float = Field(
        ..., description="The diameter of the pipette shaft.", alias="shaftDiameter"
    )
    shaft_ul_per_mm: float = Field(
        ...,
        description="The conversion rate between uL dispensed and mm of motor movement.",
        alias="shaftULperMM",
    )
    backlash_distance: float = Field(
        ...,
        description="The distance of backlash on the plunger motor.",
        alias="backlashDistance",
    )
    quirks: List[EnumSerializer[pip_types.Quirks]] = Field(
        ..., description="The list of quirks available for the loaded configuration"
    )
    tip_presence_check_distance_mm: float = Field(
        default=0,
        description="The distance the high throughput tip motors will travel to check tip status.",
        alias="tipPresenceCheckDistanceMM",
    )
    end_tip_action_retract_distance_mm: float = Field(
        default=0.0,
        description="The distance to move the head up after a tip drop or pickup.",
        alias="endTipActionRetractDistanceMM",
    )

    @field_validator("pipette_type", mode="before")
    @classmethod
    def convert_pipette_model_string(cls, v: str) -> pip_types.PipetteModelType:
        return pip_types.PipetteModelType(v)

    @field_validator("channels", mode="before")
    @classmethod
    def convert_channels(cls, v: int) -> pip_types.PipetteChannelType:
        return pip_types.PipetteChannelType(v)

    @field_validator("display_category", mode="before")
    @classmethod
    def convert_display_category(cls, v: str) -> pip_types.PipetteGenerationType:
        if not v:
            return pip_types.PipetteGenerationType.GEN1
        return pip_types.PipetteGenerationType(v)

    @field_validator("quirks", mode="before")
    @classmethod
    def convert_quirks(cls, v: List[str]) -> List[pip_types.Quirks]:
        return [pip_types.Quirks(q) for q in v]

    @field_validator("plunger_positions_configurations", mode="before")
    @classmethod
    def convert_plunger_positions(
        cls, v: Dict[str, PlungerPositions]
    ) -> Dict[pip_types.LiquidClasses, PlungerPositions]:
        return {pip_types.LiquidClasses[key]: value for key, value in v.items()}


class PipetteRowDefinition(BaseModel):
    key: str
    ordered_nozzles: List[str] = Field(..., alias="orderedNozzles")

    @field_validator("key")
    @classmethod
    def check_key_is_row(cls, v: str) -> str:
        if not ROW_NAMES.search(v):
            raise ValueError(f"{v} is not a valid row name")
        return v


class PipetteColumnDefinition(BaseModel):
    key: str
    ordered_nozzles: List[str] = Field(..., alias="orderedNozzles")

    @field_validator("key")
    @classmethod
    def check_key_is_column(cls, v: str) -> str:
        if not COLUMN_NAMES.search(v):
            raise ValueError(f"{v} is not a valid column name")
        return v


class PipetteBoundingBoxOffsetDefinition(BaseModel):
    back_left_corner: List[float] = Field(..., alias="backLeftCorner")
    front_right_corner: List[float] = Field(..., alias="frontRightCorner")


class PipetteGeometryDefinition(BaseModel):
    """The geometry properties definition of a pipette."""

    nozzle_offset: List[float] = Field(..., alias="nozzleOffset")
    path_to_3D: str = Field(
        ...,
        description="The shared data relative path to the 3D representation of the pipette model.",
        alias="pathTo3D",
    )
    nozzle_map: Dict[str, List[float]] = Field(..., alias="nozzleMap")
    pipette_bounding_box_offsets: PipetteBoundingBoxOffsetDefinition = Field(
        ..., alias="pipetteBoundingBoxOffsets"
    )
    ordered_columns: List[PipetteColumnDefinition] = Field(..., alias="orderedColumns")
    ordered_rows: List[PipetteRowDefinition] = Field(..., alias="orderedRows")
    lld_settings: Dict[str, Dict[str, float]] = Field(..., alias="lldSettings")

    @field_validator("nozzle_map", mode="before")
    @classmethod
    def check_nonempty_strings(
        cls, v: Dict[str, List[float]]
    ) -> Dict[str, List[float]]:
        # Note, the key should be able to be a regex but I think
        # we're not on a pydantic version that supports that.
        for k in v.keys():
            if not NOZZLE_MAP_NAMES.match(k):
                raise ValueError("{k} is not a valid key entry for nozzle map.")
        return v


class PipetteLiquidPropertiesDefinition(BaseModel):
    """The liquid properties definition of a pipette."""

    supported_tips: Dict[pip_types.PipetteTipType, SupportedTipsDefinition] = Field(
        ..., alias="supportedTips"
    )
    max_volume: int = Field(
        ...,
        description="The maximum supported volume of the pipette.",
        alias="maxVolume",
    )
    min_volume: float = Field(
        ...,
        description="The minimum supported volume of the pipette.",
        alias="minVolume",
    )
    default_tipracks: List[
        Annotated[str, BeforeValidator(validate_opentrons_tiprack)]
    ] = Field(
        ...,
        description="A list of default tiprack paths.",
        alias="defaultTipracks",
    )

    @field_validator("supported_tips", mode="before")
    @classmethod
    def convert_aspirate_key_to_channel_type(
        cls, v: Dict[str, SupportedTipsDefinition]
    ) -> Dict[pip_types.PipetteTipType, SupportedTipsDefinition]:
        return {pip_types.PipetteTipType[key]: value for key, value in v.items()}


class PipetteConfigurations(
    PipetteGeometryDefinition,
    PipettePhysicalPropertiesDefinition,
):
    """The full pipette configurations of a given model and version."""

    version: pip_types.PipetteVersionType = Field(
        ..., description="The version of the configuration loaded."
    )
    mount_configurations: pip_types.RobotMountConfigs = Field(
        ...,
    )
    liquid_properties: Dict[
        pip_types.LiquidClasses, PipetteLiquidPropertiesDefinition
    ] = Field(
        ..., description="A dictionary of liquid properties keyed by liquid classes."
    )

    @field_validator("liquid_properties", mode="before")
    @classmethod
    def convert_liquid_properties_key(
        cls, v: Dict[str, PipetteLiquidPropertiesDefinition]
    ) -> Dict[pip_types.LiquidClasses, PipetteLiquidPropertiesDefinition]:
        return {pip_types.LiquidClasses[key]: value for key, value in v.items()}


def liquid_class_for_volume_between_default_and_defaultlowvolume(
    volume: float,
    current_liquid_class_name: pip_types.LiquidClasses,
    available_liquid_classes: Dict[
        pip_types.LiquidClasses, PipetteLiquidPropertiesDefinition
    ],
) -> pip_types.LiquidClasses:
    """Determine the appropriate liquid class to use for a volume.

    This function has such a weird name because it is hardcoded to only use the liquid
    classes default and defaultLowVolume. It should no longer be used when those liquid
    classes change.
    """
    # For now, until we add more liquid classes, we're going to hardcode the default
    # and lowVolumeDefault liquid classes as the ones to switch between.
    has_lvd = pip_types.LiquidClasses.lowVolumeDefault in available_liquid_classes

    if not has_lvd:
        return pip_types.LiquidClasses.default
    if volume >= available_liquid_classes[pip_types.LiquidClasses.default].min_volume:
        return pip_types.LiquidClasses.default
    return pip_types.LiquidClasses.lowVolumeDefault


def default_tip_for_liquid_class(
    liquid_class_config: PipetteLiquidPropertiesDefinition,
) -> pip_types.PipetteTipType:
    """Provide a "default tip", the one with the largest volume."""
    tip_names = liquid_class_config.supported_tips.keys()
    return sorted(tip_names, key=lambda tip: tip.value)[-1]



import pytest
import typeguard

from opentrons_shared_data import module
from opentrons_shared_data.module import types

from . import list_v2_defs, list_v3_defs


@pytest.mark.parametrize("defname", list_v3_defs())
def test_v3_definitions_match_types(defname: str) -> None:
    """Test that V3 module definitions match ModuleDefinitionV3."""
    def_dict = module.load_definition("3", defname)
    typeguard.check_type(def_dict, types.ModuleDefinitionV3)


@pytest.mark.parametrize("defname", list_v2_defs())
def test_v2_definitions_match_types(defname: str) -> None:
    defdict = module.load_definition("2", defname)  # type: ignore [call-overload]
    typeguard.check_type(defdict, types.ModuleDefinitionV2)


@pytest.mark.parametrize("defname", ["magdeck", "tempdeck", "thermocycler"])
def test_v1_definitions_match_types(defname: str) -> None:
    defdict = module.load_definition("1", defname)
    typeguard.check_type(defdict, types.ModuleDefinitionV1)



import json
import pytest
from pathlib import Path
import typeguard

from opentrons_shared_data import load_shared_data
from opentrons_shared_data.protocol.types import (
    JsonProtocolV3,
    JsonProtocolV4,
    JsonProtocolV5,
)
from . import list_fixtures


@pytest.mark.parametrize("defpath", list_fixtures(3))
def test_v3_types(defpath: Path) -> None:
    defn = json.loads(load_shared_data(defpath))
    typeguard.check_type(defn, JsonProtocolV3)


@pytest.mark.parametrize("defpath", list_fixtures(4))
def test_v4_types(defpath: Path) -> None:
    defn = json.loads(load_shared_data(defpath))
    typeguard.check_type(defn, JsonProtocolV4)


@pytest.mark.parametrize("defpath", list_fixtures(5))
def test_v5_types(defpath: Path) -> None:
    defn = json.loads(load_shared_data(defpath))
    typeguard.check_type(defn, JsonProtocolV5)



import pytest
from typing import Dict, Any, cast
from opentrons_shared_data.pipette import (
    load_data,
    pipette_load_name_conversions,
    types,
)
from opentrons_shared_data.pipette.types import (
    PipetteChannelType,
    PipetteModelType,
    PipetteVersionType,
    PipetteOEMType,
    PipetteTipType,
    Quirks,
    LiquidClasses,
)


def test_load_pipette_definition() -> None:
    pipette_config_one = load_data.load_definition(
        PipetteModelType.p50,
        PipetteChannelType.SINGLE_CHANNEL,
        PipetteVersionType(major=3, minor=3),
        PipetteOEMType.OT,
    )

    assert pipette_config_one.channels == 1
    assert pipette_config_one.pipette_type.value == "p50"
    assert pipette_config_one.nozzle_offset == [-8.0, -22.0, -259.15]

    assert (
        pipette_config_one.liquid_properties[LiquidClasses.default]
        .supported_tips[PipetteTipType.t50]
        .default_aspirate_flowrate.default
        == 8.0
    )

    assert pipette_config_one.quirks == []
    pipette_config_two = load_data.load_definition(
        PipetteModelType.p50,
        PipetteChannelType.SINGLE_CHANNEL,
        PipetteVersionType(major=1, minor=0),
        PipetteOEMType.OT,
    )

    assert pipette_config_two.channels == 1
    assert pipette_config_two.pipette_type.value == "p50"
    assert pipette_config_two.nozzle_offset == [0.0, 0.0, 25.0]
    assert (
        pipette_config_two.liquid_properties[LiquidClasses.default]
        .supported_tips[PipetteTipType.t200]
        .default_aspirate_flowrate.default
        == 25.0
    )
    assert pipette_config_two.quirks == [Quirks.dropTipShake]


@pytest.mark.parametrize(
    argnames=["key_spot_check", "value_spot_check"],
    argvalues=[
        ["P1KSV10", "p1000_single_v1.0"],
        ["P1KHV33", "p1000_96_v3.3"],
        ["P20MV21", "p20_multi_v2.1"],
        ["P300MV10", "p300_multi_v1.0"],
        ["P3HMV14", "p300_multi_v1.4"],
    ],
)
def test_build_serial_number_lookup(key_spot_check: str, value_spot_check: str) -> None:
    lookup_table = load_data.load_serial_lookup_table()
    assert lookup_table[key_spot_check] == value_spot_check


@pytest.mark.parametrize(
    argnames=["pipette_model", "v1_configuration_changes"],
    argvalues=[
        ["p300_multi_v1.4", {"max_volume": 200, "min_volume": 10}],
        ["p1000_96_v3.3", {"tip_length": 40, "min_volume": 300}],
    ],
)
def test_update_pipette_configuration(
    pipette_model: str, v1_configuration_changes: Dict[str, Any]
) -> None:

    liquid_class = LiquidClasses.default
    model_name = pipette_load_name_conversions.convert_pipette_model(
        cast(types.PipetteModel, pipette_model)
    )
    base_configurations = load_data.load_definition(
        model_name.pipette_type,
        model_name.pipette_channels,
        model_name.pipette_version,
        model_name.oem_type,
    )

    updated_configurations = load_data.update_pipette_configuration(
        base_configurations, v1_configuration_changes, liquid_class
    )

    updated_configurations_dict = updated_configurations.model_dump()
    for k, v in v1_configuration_changes.items():
        if k == "tip_length":
            for i in updated_configurations_dict["liquid_properties"][liquid_class][
                "supported_tips"
            ].values():
                assert i["default_tip_length"] == v
        else:
            assert (
                updated_configurations_dict["liquid_properties"][liquid_class][k] == v
            )



"""Error categories."""

from enum import Enum
from dataclasses import dataclass
import json
from functools import lru_cache
from typing import Dict

from ..load import load_shared_data


CATEGORIES: Dict[str, Dict[str, str]] = json.loads(
    load_shared_data("errors/definitions/1/errors.json")
)["categories"]


@dataclass(frozen=True)
class ErrorCategory:
    """Individual error category data."""

    name: str
    detail: str
    code_prefix: str


def _category_from_dict_entry(entry_name: str) -> ErrorCategory:
    return ErrorCategory(
        name=entry_name,
        detail=CATEGORIES[entry_name]["detail"],
        code_prefix=CATEGORIES[entry_name]["codePrefix"],
    )


class ErrorCategories(Enum):
    """All enuemrated error categories."""

    HARDWARE_COMMUNICATION_ERROR = _category_from_dict_entry(
        "hardwareCommunicationError"
    )
    ROBOTICS_CONTROL_ERROR = _category_from_dict_entry("roboticsControlError")
    ROBOTICS_INTERACTION_ERROR = _category_from_dict_entry("roboticsInteractionError")
    GENERAL_ERROR = _category_from_dict_entry("generalError")

    @classmethod
    @lru_cache(4)
    def by_category_name(cls, category_name: str) -> "ErrorCategories":
        """Get a subsystem by its category name."""
        for entry in cls:
            if entry.value.name == category_name:
                return entry
        raise KeyError(category_name)

    @classmethod
    @lru_cache(4)
    def by_code_prefix(cls, prefix: str) -> "ErrorCategories":
        """Get an error category by its code prefix."""
        for entry in cls:
            if entry.value.code_prefix == prefix:
                return entry
        raise KeyError(prefix)
